
#include<stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>

#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/poll.h>


#include <time.h>
#include <sys/time.h>


#include "scip2.h"


#define DEBUGOUT   (0)

#define USEFILE    (1)






// ꥢݡȴϢؿ

//! ̿®ͼ
/*!
  @param aBR [in] ꤹ̿® (9600-500000bps)
  @return cfset[i|o]speedؿϤ͡
  ԤȤ B0 ֤
 */
speed_t bitrate(int aBR)
{
  switch(aBR)
    {
    case 9600:
      return B9600;

    case 19200:
      return B19200;

    case 38400:
      return B38400;

    case 57600:
      return B57600;

    case 115200:
      return B115200;

    case 230400:
      return B230400;

    case 460800:
      return B460800;

    case 500000:
      return B500000;

    default: // invalid bitrate
      return B0;
    }
}

//! ꥢݡȤ̿ѥ᡼ꤹ
/*!
  @param aFd [in] ꥢݡȤؤեǥץ
  @param aSpeed [in] ꤹ̿® (bps)
  @retval 0 
  @retval 1 
 */
int setupSerial(int aFd, int aSpeed)
{
#if 1
  struct termios term;
  int fd = aFd;

  /* Retrieve current port settings. */
  if(tcgetattr(fd, &term) < 0) {
    fprintf(stderr, "tcgetattr(3) failed on: %s", strerror(errno));
    close(fd);
    return(-1);
  }

  cfmakeraw(&term);
  cfsetispeed(&term, B19200);
  cfsetospeed(&term, B19200);
  
  /* Write the changes back to the port. */
  if(tcsetattr(fd, TCSAFLUSH, &term) < 0 ) {
    fprintf(stderr, "tcsetattr(3) failed on: %s",  strerror(errno));
    close(fd);
    return(-1);
  }
  tcflush(fd, TCIOFLUSH );
  return 1;

#else
  int speed;
  struct termios tio;

  speed=bitrate(aSpeed);
  if(speed==0)
    return 0;  // invalid bitrate

  tcgetattr(aFd,&tio);

  cfsetispeed(&tio,speed); // bitrate
  cfsetospeed(&tio,speed); // bitrate

  tio.c_cflag=(tio.c_cflag & ~CSIZE) | CS8; // data bits = 8bit

  tio.c_iflag&= ~( BRKINT | ICRNL | ISTRIP );
  tio.c_iflag&= ~ IXON;    // no XON/XOFF
  tio.c_cflag&= ~PARENB;   // no parity
  tio.c_cflag&= ~CRTSCTS;  // no CTS/RTS
  tio.c_cflag&= ~CSTOPB;   // stop bit = 1bit

  // Other
  tio.c_lflag &= ~( ISIG | ICANON | ECHO );

  // Commit
  if(tcsetattr(aFd,TCSADRAIN,&tio)==0)
    return 1;

  return 0;
#endif
}



int fileNO(FILE* fp)
{
#if USEFILE
  return fileno(fp);
#else
  return (int)fp;
#endif
}



FILE* openSerial(char *aDevName, int aSpeed)
{
  FILE* dev;

#if USEFILE
  dev=fopen(aDevName,"w+");
#else
  dev=(FILE*) open(aDevName, O_RDWR | O_SYNC | O_NONBLOCK, S_IRUSR | S_IWUSR);
#endif

  if(dev==0)
    return 0;  // invalid device file

  // setup parameters
  if(setupSerial(fileNO(dev),aSpeed))
    return dev;

  // fail
  perror("setting term parameters");

#if USEFILE
  fclose(dev);
#else
  close((int)dev);
#endif
  return 0;
}





int Scip2_Write(char* buf, int buflen, FILE* fp)
{
#if USEFILE

  int ret = fwrite(buf, sizeof(char), buflen, fp);

#if DEBUGOUT
  printf("Scip2_Write ret %d\n", ret);
#endif
  return ret;//fwrite(buf, sizeof(char), buflen, fp);

#else
  int fd = (int)fp;
  int ret = write(fd, buf, buflen);

  return ret;
#endif
}


int Scip2_Gets(char* buf, int buflenmax, FILE* fp)
{
#if USEFILE
  return (int) fgets(buf, buflenmax, fp);
#else

  int fd = (int)fp;
  int readnum = 0;
  int readadd = 0;

  //  printf("\n");

  do {
    if (readadd > 0) {
      readnum += readadd;
    }
    readadd = read(fd, &buf[readnum], 1);

#if DEBUGOUT
    printf("read %d\n", readadd);
#endif
    
    //    printf("%02x ", buf[readnum]);
    
  } while(buf[readnum] != '\n');

  if (readadd > 0) {
    readnum += readadd;
  }
  buf[readnum] = 0;
  
  return readnum;

#endif
}




void Scip2_SCIP2_0(FILE* aPort)
{

  char line[80];

#if DEBUGOUT
  printf("Scip2_SCIP2_0\n");
#endif

  Scip2_Write((char*)"SCIP2.0\n",  8, aPort);

  //echo back
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    perror("Could not change SCIP2.0");
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  //status
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    perror("Cound not change SCIP2.0");
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  //last LF
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    perror("Cound not change SCIP2.0");
    return;
  }

#if DEBUGOUT
  printf("SCIP2.0!\n");
#endif

}


//version info
void Scip2_VV(FILE* aPort)
{
  char line[80];

#if DEBUGOUT
  printf("Scip2_VV\n");
#endif
  
  Scip2_Write("VV\n",  3, aPort);
  
  //echo back
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    perror("Could not change SCIP2.0");
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  //status
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    perror("Cound not change SCIP2.0");
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  {
    int i;
    for (i = 0; i < 6; i++) {
      if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
        return;
      }
#if DEBUGOUT
      printf(line);
#endif
    }
  }

}


//light on
void Scip2_BM(FILE* aPort)
{
  char line[80];

#if DEBUGOUT
  printf("Scip2_BM\n");
#endif
  
  Scip2_Write("BM\n",  3, aPort);
  
  //echo back
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }
#if DEBUGOUT
  printf(line);
#endif

  //status
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif
}


//parameter
void Scip2_PP(FILE* aPort)
{
  char line[80];

#if DEBUGOUT
  printf("Scip2_PP\n");
#endif
  
  Scip2_Write("PP\n",  3, aPort);
  
  //echo back
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  //status
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  {
    int i;
    for (i = 0; i < 9; i++) {
      if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
        return;
      }
#if DEBUGOUT
      printf(line);
#endif
    }
  }

}

//status info
void Scip2_II(FILE* aPort)
{
  char line[80];

#if DEBUGOUT
  printf("Scip2_II\n");
#endif
  
  Scip2_Write("II\n", 3, aPort);
  
  //echo back
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  //status
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return;
  }

#if DEBUGOUT
  printf(line);
#endif

  {
    int i;
    for (i = 0; i < 7; i++) {
      if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
        return;
      }
#if DEBUGOUT
      printf(line);
#endif
    }
  }

}


int totalwrote = 0;
int totalcnt = 0;

int* Scip2_GD(FILE* aPort, int aStart, int aEnd, int aSkip, int* scan)
{
  char req[32];
  char line[80];
  int* pt = NULL;
  unsigned int i = 0;

  //tcflush(fileno(aPort), TCIOFLUSH);

  sprintf(req,"GD%04d%04d%02d\n", aStart, aEnd, aSkip);

#if DEBUGOUT
  printf("\n");
  printf("req:%s", req);
#endif

  unsigned int wrote = 0;
  while (wrote < strlen(req)) {
    

     wrote += Scip2_Write(req+wrote,  strlen(req)-wrote, aPort);


#if DEBUGOUT
     printf("wrote %d\n", wrote);
#endif
  }

  totalwrote += wrote;
#if DEBUGOUT
  printf("totalcnt %d totalwrote %d\n", totalcnt, totalwrote);
  totalcnt++;
#endif


  // echo-back
  if(Scip2_Gets(line,sizeof(line),aPort)==0) {
      return 0;
  }
#if DEBUGOUT
  printf(line);
#endif

  // status
  if(Scip2_Gets(line,sizeof(line),aPort)==0) {
      return 0;
  }
#if DEBUGOUT
  printf(line);
#endif

  // timestamp
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return 0;
  }
#if DEBUGOUT
  printf(line);
#endif

  // result
  {
    char encData[3*(1127+1)] = {0,};
    int lineEnd;


    while(1) {

      if (Scip2_Gets(line,sizeof(line),aPort) == 0) {
	printf("return NULL\n");
      }
#if DEBUGOUT
      printf(line);
#endif
      if(!strcmp("\n",line))
      {
	//end of result
	break;
      }

      lineEnd = strlen(line)-2;
      line[lineEnd] = 0x00;
      strcat(encData, line);
#if DEBUGOUT
      printf(" len %d\n", strlen(encData));
#endif
    }


#if DEBUGOUT
    printf("read end\n");
#endif

    //DECODE
    if (scan == NULL) {
      scan=(int*)malloc(sizeof(int)*((aEnd-aStart)/aSkip+1));
    }
    pt=scan;    

#if DEBUGOUT
    printf("encData len %d\n", strlen(encData));
#endif

    i = 0;
    while (i < strlen(encData)-1) {
      *pt= ((encData[i] - 0x30)<<12) + ((encData[i+1] - 0x30)<<6) + (encData[i+2] - 0x30);
      i += 3;
      pt++;
    }
  }

  return scan;
}


int* Scip2_MD(FILE* aPort, int aStart, int aEnd, int aSkip, int aSkipScan, int aSends, int* scan)
{
  char req[64];
  char line[80];
  int* pt = NULL;
  unsigned int i = 0;

  sprintf(req,"MD%04d%04d%02d%01d%02d\n", aStart, aEnd, aSkip, aSkipScan, aSends);

  Scip2_Write(req,strlen(req),aPort);


  // echo-back
  if(Scip2_Gets(line,sizeof(line),aPort)==0) {
      return 0;
  }
#if DEBUGOUT
  printf(line);
#endif

  // status
  if(Scip2_Gets(line,sizeof(line),aPort)==0) {
      return 0;
  }
#if DEBUGOUT
  printf(line);
#endif

  // timestamp
  if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
    return 0;
  }
#if DEBUGOUT
  printf(line);
#endif


  //---------------------------
  // BUG? 
  //    MD0000108001000 echo back
  //    99b             status
  //    ***             timestamp
  //---------------------------
  {
    int i = 0;
    for (i = 0; i < 3; i++) {

      if (Scip2_Gets(line, sizeof(line), aPort) == 0) {
	return 0;
      }
#if DEBUGOUT
      printf(line);
#endif
    }
  }



  // result
  {
    char encData[3*(1127+1)] = {0,};
    int lineEnd;

    while(1) {

      if (Scip2_Gets(line,sizeof(line),aPort) == 0) {
	printf("return 0\n");
      }
#if DEBUGOUT
      printf(line);
#endif
      if(!strcmp("\n",line))
      {
	//end of result
	break;
      }

      lineEnd = strlen(line)-2;
      line[lineEnd] = 0x00;
      strcat(encData, line);
#if DEBUGOUT
      printf(" len %d\n", strlen(encData));
#endif
    }



    //DECODE
    if (scan == NULL) {
      scan=(int*)malloc(sizeof(int)*((aEnd-aStart)/aSkip+1));
    }
    pt=scan;    

    i = 0;
    while (i < strlen(encData)-1) {
      *pt= ((encData[i] - 0x30)<<12) + ((encData[i+1] - 0x30)<<6) + (encData[i+2] - 0x30);
      i += 3;
      pt++;
    }
  }


  return scan;
}




FILE* Scip2_Start(char* device)
{
  FILE* dev = openSerial(device, 19200);
  if(dev == NULL)
    {
      perror("Could not open serial port");
      exit(1);
    }

#if 0
  Scip2_SCIP2_0(dev);
#endif
  
  Scip2_VV(dev);
  Scip2_BM(dev);
  Scip2_PP(dev);
  Scip2_II(dev);

  return dev;
}


void Scip2_End(FILE* dev)
{
  if (dev) 
  {
    fclose(dev);
    dev = NULL;
  }
}








